#![no_std]
#![allow(dead_code, non_camel_case_types)]

use core::panic::PanicInfo;
#[panic_handler]
fn panic(_info: &PanicInfo) -> ! {
    loop {}
}

mod zephyr_sys {
    // Module containing Rust definitions of common C types
    use super::cty;

    // Module containing the output of Kconfig and DTC
    pub mod config;

    pub const GPIO_DIR_OUT: i32 = 1 << 0;

    #[repr(C)]
    pub struct device_config {
        name: *const cty::c_char,
        init: Option<unsafe extern "C" fn (foo: *mut device) -> cty::c_int>,
        config_info: *const cty::c_void,
    }

    #[repr(C)]
    pub struct device {
        pub config: *mut device_config,
        pub driver_api: *mut cty::c_void,
        pub driver_data: *mut cty::c_void,
    }

    // A C #define used inside Zephyr's gpio library
    const GPIO_ACCESS_BY_PIN: cty::c_int = 0;

    struct _snode {
        next: *mut _snode,
    }

    type sys_snode_t = _snode;

    type gpio_callback_handler_t = 
        Option<unsafe extern "C" fn(port: *mut device, cb: *mut gpio_callback, 
                                    pins: u32)>;

    // This union is defined inside the struct gpio_callback 
    // and not given a name in Zephyr.
    #[repr(C)]
    union gpio_callback_anon_union {
        pin_mask: u32,
        pin: u32,
    }

    #[repr(C)]
    struct gpio_callback {
        node: sys_snode_t,
        handler: gpio_callback_handler_t,
        my_field: gpio_callback_anon_union,
    }

    type gpio_config_t = 
        Option<extern "C" fn(port: *mut device, access_op: cty::c_int, pin: u32, 
                             flags: cty::c_int) -> cty::c_int>;
    type gpio_write_t = 
        Option<extern "C" fn(port: *mut device, access_op: cty::c_int, pin: u32, 
                             value: u32) -> cty::c_int>;
    type gpio_read_t = 
        Option<extern "C" fn(port: *mut device, access_op: cty::c_int, pin: u32, 
                             value: *mut u32) -> cty::c_int>;
    type gpio_manage_callback_t = 
        Option<extern "C" fn(port: *mut device, callback: *mut gpio_callback, 
                             set: bool) -> cty::c_int>;
    type gpio_enable_callback_t = 
        Option<extern "C" fn(port: *mut device, access_op: cty::c_int, pin: u32)
                             -> cty::c_int>;
    type gpio_disable_callback_t = 
        Option<extern "C" fn(port: *mut device, access_op: cty::c_int, pin: u32)
                             -> cty::c_int>;
    type gpio_api_get_pending_int = Option<extern "C" fn(dev: *mut device) -> u32>;

    #[repr(C)]
    struct gpio_driver_api {
        config: gpio_config_t,
        write: gpio_write_t,
        read: gpio_read_t,
        manage_callback: gpio_manage_callback_t,
        disable_callback: gpio_disable_callback_t,
        get_pending_int: gpio_api_get_pending_int,
    }

    fn _impl_gpio_config(port: *mut device, access_op: cty::c_int, pin: u32, 
                         flags: cty::c_int) -> cty::c_int 
    {
        unsafe {
            let api = (*port).driver_api as *const gpio_driver_api;
            return ((*api).config)(port, access_op, pin, flags);
        }
    }

    // This function is static inline in C, so we define it here
    #[no_mangle]
    pub fn gpio_pin_configure(port: *mut device, pin: u32, flags: cty::c_int) 
                              -> cty::c_int 
    {
        return _impl_gpio_config(port, GPIO_ACCESS_BY_PIN, pin, flags);
    }

    fn _impl_gpio_write(port: *mut device, access_op: cty::c_int, pin: u32, value: u32) 
                        -> cty::c_int 
    {
        unsafe {
            let api = (*port).driver_api as *const gpio_driver_api;
            return ((*api).write)(port, access_op, pin, value);
        }
    }

    // This function is static inline in C, so we define it here
    #[no_mangle]
    pub fn gpio_pin_write(port: *mut device, pin: u32, value: u32) -> cty::c_int 
    {
        return _impl_gpio_write(port, GPIO_ACCESS_BY_PIN, pin, value);
    }

    // Bindings to external C functions
    extern "C" {
        pub fn _impl_device_get_binding(name: *const cty::c_char) -> *mut device;

        pub fn _impl_k_sleep(duration: i32);
    }
}

const LED_PORT: *const cty::c_char = zephyr_sys::dts::LED0_GPIO_CONTROLLER;
const LED: u32 = zephyr_sys::dts::LED0_GPIO_PIN;

// 1000 msec = 1 sec
const SLEEP_TIME: i32 = 1000;

#[no_mangle]
extern "C" fn main() {
    unsafe {
        let mut cnt = 0;
        let dev = zephyr_sys::_impl_device_get_binding(LED_PORT);

        zephyr_sys::gpio_pin_configure(dev, LED, zephyr_sys::GPIO_DIR_OUT);

        loop {
            zephyr_sys::gpio_pin_write(dev, LED, cnt % 2);
            cnt += 1;
            zephyr_sys::_impl_k_sleep(SLEEP_TIME);
        }
    }
}
